module Show

# See https://github.com/JuliaLang/julia/pull/36264

# https://github.com/JuliaLang/julia/blob/293e8132c067a1ef2b772d773530d4d75592d435/base/show.jl

using Base: @nospecialize, inferencebarrier, show_circular

function _show_without_type_parameters(io::IO, @nospecialize(x))
    t = typeof(x)
    print(io, t.name.name)
    print(io, '(')
    show_fields(io, x)
    print(io, ')')
end

function _show_default(io::IO, @nospecialize(x))
    t = typeof(x)
    show(io, inferencebarrier(t))
    print(io, '(')
    show_fields(io, x)
    print(io, ')')
end

function show_fields(io::IO, @nospecialize(x))
    t = typeof(x)
    nf = nfields(x)
    nb = sizeof(x)
    if nf != 0 || nb == 0
        if !show_circular(io, x)
            recur_io = IOContext(io, Pair{Symbol,Any}(:SHOWN_SET, x),
                                 Pair{Symbol,Any}(:typeinfo, Any))
            for i in 1:nf
                f = fieldname(t, i)
                if !isdefined(x, f)
                    print(io, undef_ref_str)
                else
                    show(recur_io, getfield(x, i))
                end
                if i < nf
                    print(io, ", ")
                end
            end
        end
    else
        print(io, "0x")
        r = Ref(x)
        GC.@preserve r begin
            p = unsafe_convert(Ptr{Cvoid}, r)
            for i in (nb - 1):-1:0
                print(io, string(unsafe_load(convert(Ptr{UInt8}, p + i)), base = 16, pad = 2))
            end
        end
    end
end

end